/**
    Battery analyzer
    (c) 2016-2017 F. Štefanec
    This project is provided WITHOUT ANY WARRANTY, use at your own risk.
**/

/// VERSION 1.4.0


/** You can find all settings in main.h **/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "main.h"

#include "string_table.h"

#include "lcd.h" // LCD library

#include "digits.h" // Number to string conversion
#include "lcd_customchars.h" // Custom character definitions

#include "menu.h" // menu functions

#include "global_defines.h"


/** Timer 1 Compare ISR - time counter **/
ISR(TIMER1_COMPA_vect)
{
    if (stat & stat_disch_enabled)
    {
        stat |= stat_display_prog;

        seconds++;
        if (seconds>=60)
        {
            seconds = 0;
            minutes++;
        }
        if (minutes>=60)
        {
            minutes = 0;
            hours++;
        }
        if (hours==255)
        {
            stat &= ~stat_disch_enabled;
            stat |= stat_term_wait;
        }
    }

    if (stat & stat_ccload_enabled)
    {
        stat |= stat_ccload_display;
    }

    // battery icon display

    batt_icon++;
    if (batt_icon > 4) batt_icon = 0;

}

/** AD conversion finished ISR **/
ISR(ADC_vect)
{
    if (ad_throwaway == 0) /// throw away a few measurements after a MUX change
    {
        /// ADC (pre)averaging (x32)

        ad_avg_buf[ad_avg_buf_ptr] = ADC;
        ad_avg_buf_ptr++;

        /// Averaging (x32) - total x1024 - oversampling is used to gain higher resolution
        if(ad_avg_buf_ptr>=ad_preavg_count)
        {
            ad_avg_buf_ptr = 0;

            uint8_t array_ptr, new_admux;

            uint16_t ADC_VALUE = 0;

            uint8_t i;

            for (i=0; i<ad_preavg_count; i++)
            {
                ADC_VALUE += ad_avg_buf[i];
            }

            ADC_VALUE /= ad_preavg_count;

            array_ptr = ad_buf_ptr;

            new_admux = ADMUX & ~((1<<MUX0) | (1<<MUX1) | (1<<MUX2) | (1<<MUX3));

            if(array_ptr<32) /// compute battery voltage (32x ADC measurements divided by 16)
            {
                volt_buf[array_ptr] = ADC_VALUE;

                uint8_t j = 0;
                uint16_t adc_volt_temp = 0;

                for (j=0; j<32; j++)
                {
                    adc_volt_temp += volt_buf[j];
                }

                adc_voltage = adc_volt_temp;

                if (offset_comp != 0)
                {
                    uint8_t abs_val = (offset_comp<0) ? (uint8_t)(-1*offset_comp) : (uint8_t)(offset_comp);

                    if ((offset_comp != (int8_t)abs_val) && (adc_voltage > abs_val))
                    {
                        adc_voltage -= abs_val;
                    }
                    else
                    {
                        adc_voltage += offset_comp;
                    }

                }

                adc_voltage /= 16;

                if (adc_voltage > 2500) adc_voltage = 0; /// in case of math error

                if (offset_comp > 0 && adc_voltage <= 1+offset_comp/16) adc_voltage = 0; /// measured offset voltage = measured practically zero
            }

            if (array_ptr==32) supp_voltage = ADC_VALUE;
            if (array_ptr==33) temperature_value = ADC_VALUE;

            ad_buf_ptr++;
            if (ad_buf_ptr==34) ad_buf_ptr = 0;

            if (ad_buf_ptr==0)
            {
                new_admux |= MUX_BATT_VOLTAGE;
                ad_throwaway = 5;
                ADMUX = new_admux;
            }
            if (ad_buf_ptr==32)
            {
                new_admux |= MUX_SUPPLY_VOLTAGE;
                ad_throwaway = 5;
                ADMUX = new_admux;
            }
            if (ad_buf_ptr==33)
            {
                new_admux |= MUX_TEMPERATURE;
                ad_throwaway = 5;
                ADMUX = new_admux;
            }
        }
    }
    else
    {
        ad_throwaway--;
        _delay_us(1);
    }

}

/** Timer 0 compare ISR **/
/** f = 2 kHz **/

ISR(TIMER0_COMPA_vect)
{
    tim0_sw_pre++;
    if (tim0_sw_pre > 7)
    {
        /// z stores the port input, y stores the button values

        z = ~BTN_PIN;
        z &= ((1<<BTN_ESC_PIN) | (1<<BTN_OK_PIN) | (1<<BTN_DOWN_PIN) | (1<<BTN_UP_PIN));
        y = 0;

             if ((z & (1<<BTN_ESC_PIN)) != 0) y |= BTN_ESC;
        else if ((z & (1<<BTN_UP_PIN)) != 0) y |= BTN_UP;
        else if ((z & (1<<BTN_DOWN_PIN)) != 0) y |= BTN_DOWN;
        else if ((z & (1<<BTN_OK_PIN)) != 0) y |= BTN_OK;

        /// debounce counter

        if (btn_debounce == 0)
        {
            buttons = y;
            btn_debounce = DEBOUNCE_VAL;
        }
        else btn_debounce--;

        /// beeper counter

        if (buttons>0 && (buttons!=buttons_old_pin) && beep_counter == 0)
        {
            beep_counter = BEEP_LEN_BUTTON;
        }

        buttons_old_pin = buttons;

        /// button repeat counter

        if (y==0)
        {
            button_repeat_counter=0;
        }
        else
        {
            button_repeat_counter++;
            if (button_repeat_counter == REPEAT_DELAY)
            {
                buttons_old = 0;
                button_repeat_counter -= REPEAT_PERIOD;
                beep_counter = BEEP_LEN_REPEAT;
            }
        }

        /// termination voltage check

        if ((stat & stat_disch_enabled)>0 && ((adc_voltage <= term_voltage)))
        {
            terminate_discharge();
            stat |= stat_term_wait;
        }

        /// beep counter decrement

        if (beep_counter>0)
        {
           beep_counter--;
        }

        tim0_sw_pre=0;
    }

    /// beep
    if (beep_counter==0) BUZZER_PORT &= ~(1<<BUZZER_PIN);
    else if (settings & settings_sound_on) BUZZER_PORT ^= (1<<BUZZER_PIN);

    _delay_us(1);


    /// start A->D conversion
    ADCSRA |= (1<<ADSC);


    /// power LED control
    tim0_blink_counter++;
    if (tim0_blink_counter >= 1000) tim0_blink_counter = 0;
    tim0_even ++;
    if (tim0_even >= 2) tim0_even = 0;

    if (led_blink_time <= tim0_blink_counter)
    {
        if (led_state1 == 0 || (led_state1 == 1 && tim0_even == 1)) LED_PORT &= ~(1<<LED_PIN);
        else LED_PORT |= (1<<LED_PIN);
    }
    else
    {
        if (led_state2 == 0 || (led_state2 == 1 && tim0_even == 1)) LED_PORT &= ~(1<<LED_PIN);
        else LED_PORT |= (1<<LED_PIN);
    }


    /// temperature based fan control
    #ifdef LM35_INSTALLED
    if (temperature_value < TEMP_AD_FAN_STOP) fan_control &= ~2;
    if (temperature_value > TEMP_AD_FAN_START) fan_control |= 2;

    if (temperature_value > TEMP_AD_FAULT)
    {
        cli();
        disable_wdt();
        deconf_regs();
        FAN_PORT |= (1<<FAN_PIN);
        print_temp_error();
        halt_error();
    }

    if ((fan_control & 2) != 0) FAN_PORT |= (1<<FAN_PIN); else FAN_PORT &= ~(1<<FAN_PIN);
    #else
    if ((fan_control & 1) != 0) FAN_PORT |= (1<<FAN_PIN); else FAN_PORT &= ~(1<<FAN_PIN);
    #endif

}

void clear_time()
{
    hours = 0;
    minutes = 0;
    seconds = 0;
}

void enable_wdt()
{
    asm("WDR");
    WDTCSR |= (1<<WDCE);
    WDTCSR |= (1<<WDP2) | (1<<WDP0) | (1<<WDE); /*0.5s timeout and WD enabled*/
}

void disable_wdt()
{
    asm("WDR");
    WDTCSR |= (1<<WDCE);
    WDTCSR = 0;
}

/** print something from progmem on LCD **/
void lcd_print_pmem(const char * data, uint8_t len)
{
    int n;
    for (n=0;n<len;n++)
        buf[n] = pgm_read_byte_near(data+n);
    lcd_sendstring(buf, len);
}

/** set display cursor position & print something from progmem on LCD **/
void lcd_print_pmem_setpos(const char * data, uint8_t len, uint8_t column, uint8_t line)
{
    int n;
    for (n=0;n<len;n++)
        buf[n] = pgm_read_byte_near(data+n);
    lcd_setcursor(column, line);
    lcd_sendstring(buf, len);
}

void enable_1s_timer()
{
    cli();
    TCNT1 = 0;
    OCR1A = TIME_PRESCALER_1000ms;
    TIMSK1 |= (1<<OCIE1A); // enable CTC interrupt
    sei();
}

/** set "1s" timer to run 4x faster **/
void fast_1s_timer()
{
    cli();
    TCNT1 = 0; // reset timer
    OCR1A = TIME_PRESCALER_250ms;
    TIMSK1 |= (1<<OCIE1A); // enable CTC interrupt
    sei();
}



void disable_1s_timer()
{
    cli();
    TIMSK1 &= ~(1<<OCIE1A);
    sei();
}

void terminate_discharge()
{
    /// read voltage before termination
    end_voltage = adc_voltage;

    /// reset and disable timers
    set_current(0);
    disable_1s_timer();
    disable_wdt();

    calculate_mAh();
    stat &= ~stat_disch_enabled;
    lcd_clear();

    /// print results on display
    print_discharge_result();

    /** beep for 1s **/

    beep_counter = BEEP_LEN_FINISHED;

    /** set current to zero **/
    set_current(0);
}

void print_discharge_result()
{
    lcd_setcursor(0,0);
    lcd_print_pmem(str_terminated, 13);
    voltage_string(buf, end_voltage);
    lcd_sendstring(buf, 6);

    if (display_mode == 0)
    {
        lcd_setcursor(0,1); // print mAh and discharge current
        mAh_string(buf);
        lcd_sendstring(buf, 9);
        lcd_print_pmem(str_mAh_fin, 6);
        current_string(buf, pwm_current);
        lcd_sendstring(buf, 5);
    }
    else
    {
        lcd_setcursor(0,1); // print mAh and discharge current
        mWh_string(buf);
        lcd_sendstring(buf, 9);
        lcd_print_pmem(str_mWh_fin, 6);
        current_string(buf, pwm_current);
        lcd_sendstring(buf, 5);
    }

    lcd_print_pmem_setpos(elapsed, 9, 0, 2); // elapsed time
    time_string(buf);
    lcd_sendstring(buf, 9);

    lcd_print_pmem_setpos(str_esc_exit, 20, 0, 3); // continue prompt
    stat &= ~stat_disch_enabled;
    stat |= stat_term_wait;
}

void mAh_string(char* buf)
{
    buf[0] = dig32_7(mAh100);
    buf[1] = dig32_6(mAh100);
    buf[2] = dig32_5(mAh100);
    buf[3] = dig32_4(mAh100);
    buf[4] = dig32_3(mAh100);
    buf[5] = dig32_2(mAh100);
    buf[6] = '.';
    buf[7] = dig32_1(mAh100);
    buf[8] = dig32_0(mAh100);
}

void mWh_string(char* buf)
{
    unsigned long long mWh = mWsx10 / 36000; // get mWh

    buf[0] = dig32_7(mWh);
    buf[1] = dig32_6(mWh);
    buf[2] = dig32_5(mWh);
    buf[3] = dig32_4(mWh);
    buf[4] = dig32_3(mWh);
    buf[5] = dig32_2(mWh);
    buf[6] = dig32_1(mWh);
    buf[7] = dig32_0(mWh);
    buf[8] = ' ';
}

void watt_string(char* buf, uint16_t voltage, uint8_t current)
{
    unsigned long long wattage = ((unsigned long long)voltage)*((unsigned long long)current);
    /*if (voltage>=MIN_VOLTAGE)
    {*/
        buf[0] = dig32_5(wattage);
        buf[1] = dig32_4(wattage);
        buf[2] = '.';
        buf[3] = dig32_3(wattage);
        buf[4] = dig32_2(wattage);
        buf[5] = dig32_1(wattage);
        buf[6] = dig32_0(wattage);
    /*}
    else
    {
        buf[0] = '-';
        buf[1] = '-';
        buf[2] = '-';
        buf[3] = '-';
        buf[4] = '-';
        buf[5] = '-';
        buf[6] = '-';
    }*/
    buf[7] = 'W';
}


void voltage_string(char* buf, uint16_t voltage)
{
    if (voltage>=10 || 1) /// temporarily always on for debug purposes
    {
        buf[0] = dig16_3(voltage);
        buf[1] = dig16_2(voltage);
        buf[2] = '.';
        buf[3] = dig16_1(voltage);
        buf[4] = dig16_0(voltage);
    }
    else
    {
        buf[0] = '-';
        buf[1] = '-';
        buf[2] = '-';
        buf[3] = '-';
        buf[4] = '-';
    }
	buf[5] = 'V';
}

void current_string(char* buf, uint8_t current)
{
    buf[0] = dig8_2(current);
	buf[1] = '.';
	buf[2] = dig8_1(current);
	buf[3] = dig8_0(current);
	buf[4] = 'A';
}

void time_string(char* buf)
{
    buf[0] = dig8_2(hours);
	buf[1] = dig8_1(hours);
	buf[2] = dig8_0(hours);
	buf[3] = ':';
	buf[4] = dig8_1(minutes);
	buf[5] = dig8_0(minutes);
	buf[6] = ':';
	buf[7] = dig8_1(seconds);
	buf[8] = dig8_0(seconds);
}

void display_prog() /// display discharge progress
{

    lcd_print_pmem_setpos(elapsed, 9, 0, 0); // 1st line - elapsed time
	time_string(buf);
	lcd_sendstring(buf, 9);

    lcd_print_pmem_setpos(str_voltage, 9, 0, 1); // 2nd line - voltage from ADC
	voltage_string(buf, adc_voltage);
	lcd_sendstring(buf, 6);

	if (display_mode == 0)
	{

        lcd_print_pmem_setpos(str_current, 9, 0, 2); // 3rd line - current
        current_string(buf, pwm_current);
        lcd_sendstring(buf, 5);
        lcd_printspaces(3);

        lcd_print_pmem_setpos(str_mAh, 5, 0, 3); // 4th line - mAh
        mAh_string(buf);
        lcd_sendstring(buf, 9);
    }

    else
    {
        lcd_print_pmem_setpos(str_power, 9, 0, 2); // 3rd line - power
        watt_string(disch_power_str, adc_voltage, pwm_current);
        lcd_sendstring(disch_power_str, 8);

        lcd_print_pmem_setpos(str_mWh, 5, 0, 3); // 4th line - mAh
        mWh_string(buf);
        lcd_sendstring(buf, 9);


    }
    // 4th line, 20th col - discharging battery icon

    lcd_setcursor(19,3);
    lcd_writedata();
    lcd_send(batt_icon);
}


void display_ccload() /// display voltage, current
{
    lcd_print_pmem_setpos(str_voltage, 9, 0, 0); // 1st line - voltage from ADC
    disp_voltage = adc_voltage;

	voltage_string(buf, disp_voltage);
	lcd_sendstring(buf, 6);

	if(disp_voltage < 80)
	{
        lcd_print_pmem(str_low,5);
	}
	else lcd_printspaces(5);

    lcd_print_pmem_setpos(str_current, 9, 0, 1); // 2nd line - current
	current_string(buf, pwm_current);
	lcd_sendstring(buf, 5);

	lcd_print_pmem_setpos(str_power, 7, 0, 2);

	lcd_sendstring(disch_power_str, 8);
}

void calculate_mAh()
{
    elapsed_secs = ((hours*60U+minutes)*60UL)+seconds;

    mAs = pwm_current;
    mAs *= 10ULL;
    mAs *= (unsigned long long)elapsed_secs;
    mAh100 = mAs / 36ULL;

}

void calculate_mWsx10()
{
    unsigned long long add_val;
    add_val = (unsigned long long)(adc_voltage);
    add_val *= (unsigned long long)(pwm_current);
    mWsx10 += add_val;
}

void rin_test() /// try to measure internal resistance
{
    led_blink_time = 500;
    led_state2 = 1;

    ad_preavg_count = RIN_PREAVG_COUNT; // averaging only 8x32 measurements (256) instead of 32x32 (1024) for speed

    lcd_clear();
    lcd_print_pmem(str_rin_wait, 20);

    i1 = pgm_read_byte_near(rin_current1+rin_current_choice);
    i2 = pgm_read_byte_near(rin_current2+rin_current_choice);

    uint16_t v1avg=0, v2avg=0;

    set_current(i1);
    for (x=0; x<200; x++) _delay_ms(5); // 1000ms wait loop.

    uint8_t runs;
    for (runs=0; runs<8; runs++)
    {

        for (x=0; x<100; x++) _delay_ms(5); // 500ms wait loop
        v1 = adc_voltage;
        v1avg += v1;
        if (v1<80 || v1>2100)
        {
            v1avg = 0;
            v2avg = 0;
            break;
        }

        set_current(i2);

        for (x=0; x<100; x++) _delay_ms(5); // 500ms wait loop
        v2 = adc_voltage;
        v2avg += v2;
        if (v2<80 || v2>2100)
        {
            v1avg = 0;
            v2avg = 0;
            break;
        }

        set_current(i1);

    }
    set_current(0);
    v1avg /= 8;
    v2avg /= 8;
    v1 = v1avg;
    v2 = v2avg;

    if (v1avg < 10 || v2avg < 10)
    {

        lcd_clear();
        lcd_print_pmem_setpos(str_error, 5, 0, 0);
        lcd_print_pmem_setpos(str_rin_psu_weak, 16, 0, 1);
        lcd_print_pmem_setpos(str_esc_exit, 20, 0, 3);

    } // supply too weak.
    else
    {
        lcd_clear();

        if (v1>=v2)
        {
            rin = v1;
            rin -= v2;
        }
        else
        {
            rin = v2;
            rin -= v1;
        }

        rin = (((unsigned long long)rin*100)/(i2-i1));

        lcd_print_pmem_setpos(str_rin_resistance, 12, 0, 0);

        buf[0] = dig16_3(rin);
        buf[1] = dig16_2(rin);
        buf[2] = '.';
        buf[3] = dig16_1(rin);
        buf[4] = dig16_0(rin);
        buf[5] = OMEGA_CHAR;

        if (v2>v1) lcd_sendstring("-", 1);
        lcd_sendstring(buf, 6);

        lcd_setcursor(0,1);

        lcd_print_pmem(str_rin_v1, 3);
        voltage_string(buf, v1);
        lcd_sendstring(buf, 6);
        lcd_print_pmem(str_rin_v2, 4);
        voltage_string(buf, v2);
        lcd_sendstring(buf, 6);

        lcd_print_pmem_setpos(str_rin_i1, 3, 0, 2);
        current_string(buf, i1);
        buf[5] = ' ';
        lcd_sendstring(buf, 6);
        lcd_print_pmem(str_rin_i2, 4);
        current_string(buf, i2);
        lcd_sendstring(buf, 5);
        lcd_print_pmem_setpos(str_esc_exit, 20, 0, 3);

    }

    led_blink_time = 0;

    stat &= ~stat_rin_test;
    stat |= stat_term_wait;
    ad_preavg_count = CAP_PREAVG_COUNT;
}

uint8_t get_buttons()
{
    uint8_t btns;
    btns = (buttons ^ buttons_old) & buttons; // buttons - pressed down
    buttons_old = buttons;
    return btns;
}


int main(void)
{
    cli();

    /** set up ports **/
    PORTD = 0;
    PORTC = 0;
    PORTB = 0;
    DDRD = 0;
    DDRC = 0;
    DDRB = 0;

    BTN_PORT |= (1<<BTN_ESC_PIN) | (1<<BTN_DOWN_PIN) | (1<<BTN_UP_PIN) | (1<<BTN_OK_PIN);

    FAN_DDR |= (1<<FAN_PIN);

    BUZZER_DDR |= (1<<BUZZER_PIN);

    LCD_DDDR |= (1<<LCD_D7) | (1<<LCD_D6) | (1<<LCD_D5) | (1<<LCD_D4) | (1<<PD0);
    LCD_CDDR |= (1<<LCD_E) | (1<<LCD_RS);

    INHIBIT_PORT |= (1<<INHIBIT_PIN);
    INHIBIT_DDR |= (1<<INHIBIT_PIN);

    DDRB |= (1<<PB3); // PWM pin

    DDRC |= (1<<PC5); // power LED pin
    PORTC = 0b11100000;

    load_eeprom();

    ad_preavg_count = CAP_PREAVG_COUNT;

    /** initialize and setup display **/
    lcd_init();
    lcd_send(LCD_FUNCSET | LCD_FUNCSET_2LINES);
    lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY);
    lcd_send(LCD_CLEAR);
    _delay_ms(5);
    lcd_send(LCD_ENTRYMODE | LCD_ENTRYMODE_INCREMENT);

    /** if the watchdog reset flag is set **/
    if (MCUSR & (1<<WDRF))
    {
        set_current(0);
        disable_1s_timer();
        lcd_clear();
        disable_wdt();
        lcd_print_pmem_setpos(str_em_shdn, 20, 0, 0);
        lcd_print_pmem_setpos(str_em_shdn_wd0, 20, 0, 1);
        lcd_print_pmem_setpos(str_em_shdn_wd1, 20, 0, 2);
        MCUSR &= ~(1<<WDRF);
        halt_error();
    }

    /** load custom characters into CGRAM **/
    lcd_customchar(0, customchar1);
    lcd_customchar(1, customchar2);
    lcd_customchar(2, customchar3);
    lcd_customchar(3, customchar4);
    lcd_customchar(4, customchar5);
    lcd_customchar(5, customchar6);

    /** print welcome string **/

    lcd_print_pmem_setpos(str_welcome1, 20, 0, 0);
    lcd_print_pmem_setpos(str_welcome2, 20, 0, 1);
    lcd_print_pmem_setpos(str_firmware_version, 20, 0, 3);

    /** ~2 sec delay **/
    for (z=0; z<250; z++) _delay_ms(8);


    /** welcome beep & screen erase **/
    beep_counter = BEEP_LEN_START;
    lcd_clear();

    /** set up timers **/

    TCCR0B = (1<<CS01) | (1<<CS00); /// 1:64 prescaler selected for Timer 0
    TCCR0A = (1<<WGM01); /// Timer 0 set to CTC mode
    OCR0A = 124; /// 125 clocks for timer 0, 16M:64:125 ~ 2 kHz
    TCCR1B = (1<<WGM12) | (1<<CS12); /// Timer 1 prescaler set to 256, CTC mode selected
    OCR1A = 62499; /// compare value of Timer1 = 62499 (16M/256/62500 = 1, 1 Hz tick frequency)
    /// Timer 0 & Timer 1 compare interrupt masks enabled
    TIMSK0 = (1<<OCIE0A);
    TIMSK1 = (1<<OCIE1A);

    OCR2A = 0; /// Set PWM value to 0
     /// enable Phase Correct waveform generation mode, non inverting, no clock prescaling (1:1 clock -> 62.5 kHz PWM at 16 MHz CPU freq)
    TCCR2A = (1<<COM2A1) | (1<<WGM20);
    TCCR2B = (1<<CS20);

    /** set up ADC **/

    ADMUX = (1<<REFS0) | (1<<REFS1); /// select internal 1.1V ADC reference & ADC0 as input
    ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0) | (1<<ADIE); /// enable ADC, ADC interrupt and set prescaler to :64
    ADCSRA |= (1<<ADSC); /// start conversion

    LED_DDR |= (1<<LED_PIN);
    led_state1 = 2;
    led_state2 = 0;
    led_blink_time = 0;


    sei();  /// enable all interruppts

    _delay_ms(5);

    /** service menu **/

    /*if (get_buttons() & (BTN_OK | BTN_DOWN))
    {
        service_menu();
    }*/

    lcd_clear();

    /** print menu **/

    stat |= stat_inside_menu;
    stat2 |= stat2_redraw_menu;

    lcd_print_pmem_setpos(str_select0, 10, 5, 0);

    set_current(0);

    while(1)
    {
        if (stat2 & stat2_redraw_menu)
        {
            lcd_print_pmem_setpos(str_select0, 10, 5, 0);
            lcd_setcursor(2,1);
            strcpy_P(buf, str_menu_items[menu_pos2+0]);
            lcd_sendstring(buf, 18);
            lcd_setcursor(2,2);
            strcpy_P(buf, str_menu_items[menu_pos2+1]);
            lcd_sendstring(buf, 18);
            lcd_setcursor(2,3);
            strcpy_P(buf, str_menu_items[menu_pos2+2]);
            lcd_sendstring(buf, 18);

            stat2 &= ~stat2_redraw_menu;
        }

        if (stat & stat_inside_menu)
        {
            led_blink_time = 0;
            if (OCR2A != 0) set_current(0);
            disp_menu();
        }

        if (stat & stat_display_prog)
        {
            led_blink_time = 200;
            led_state2 = 0;
            calculate_mAh();
            calculate_mWsx10();
            display_prog();
            stat &= ~(stat_display_prog);
        }

        if (stat & stat_ccload_display)
        {
            led_blink_time = 200;
            led_state2 = 0;
            display_ccload();
            stat &= ~(stat_ccload_display);
        }

        if (stat & stat_term_wait)
        {
            led_blink_time = 500;
            led_state2 = 0;
            buttons_pressed = get_buttons();
            if(buttons_pressed & BTN_ESC)
            {
                disable_wdt();
                lcd_clear();
                print_menu(str_menu_items, menu_pos2);
                stat &= ~(stat_term_wait);
                stat |= stat_inside_menu;
            }

            if((buttons_pressed & BTN_OK || buttons_pressed & BTN_UP || buttons_pressed & BTN_DOWN) && (last_option == 0))
            {
                display_mode = display_mode == 1 ? 0 : 1;
                print_discharge_result();
            }
        }

        if (stat & (stat_ccload_enabled | stat_disch_enabled))
        {

            watt_string(disch_power_str, disp_voltage, pwm_current);

            buttons_pressed = get_buttons();
            if (buttons_pressed & BTN_ESC)
            {
                set_current(0);
                if (stat & stat_ccload_enabled)
                {
                    stat &= ~stat_ccload_enabled;
                    disable_wdt();
                    escape_to_menu();
                }
                else if (stat & stat_disch_enabled)
                {
                    stat &= ~stat_disch_enabled;
                    stat |= stat_term_wait;
                    terminate_discharge();
                }
            }

            if((buttons_pressed & BTN_OK) && (stat & stat_ccload_enabled))
            {
                display_mode = display_mode == 1 ? 0 : 1;
            }

            else if ((buttons_pressed & (BTN_UP | BTN_DOWN)) && (stat & stat_ccload_enabled))
            {
                if ((buttons_pressed & BTN_UP) && pwm_current < 255)
                {
                    set_current(0);
                    pwm_current++;
                    set_current(pwm_current);
                }
                if ((buttons_pressed & BTN_DOWN) && pwm_current > 1)
                {
                    set_current(0);
                    pwm_current--;
                    set_current(pwm_current);
                }
                display_ccload();
            }

            if ((stat & stat_disch_enabled) && (buttons_pressed & BTN_OK || buttons_pressed & BTN_UP || buttons_pressed & BTN_DOWN))
            {
                display_mode = display_mode == 1 ? 0 : 1;
                display_prog();
            }

        }
        asm("WDR"); // reset watchdog
    }

    return 0;

}



void set_current(uint8_t current)
{
    OCR2A = current;
    if (current==0)
    {
        INHIBIT_PORT |= (1<<INHIBIT_PIN);
        fan_control &= ~1;
    }
    else
    {
        INHIBIT_PORT &= ~(1<<INHIBIT_PIN);
        fan_control |= 1;
    }

}


void halt_error()
{

    cli();
    set_current(0);
    disable_1s_timer();
    disable_wdt();
    disable_1s_timer();
    while(1);
}



/// ----------------------------------------- MENU FUNCTIONS ----------------------------------------------

void service_menu()
{
    lcd_clear();
    uint8_t settings_menu_pos=0, settings_menu_pos_old=0;

    lcd_print_pmem_setpos(str_service_menu_title, 20, 0, 0);
    lcd_print_pmem_setpos(str_srv_menu_items[0], 14, 2, 1);
    lcd_print_pmem_setpos(str_srv_menu_items[1], 14, 2, 2);

    lcd_setcursor(0,1); lcd_writedata(); lcd_send('>');

    while(1)
    {
        buttons_pressed = get_buttons();
        if ((buttons_pressed & BTN_UP) && (settings_menu_pos>0)) settings_menu_pos--;
        if ((buttons_pressed & BTN_DOWN) && (settings_menu_pos<1)) settings_menu_pos++;
        if (settings_menu_pos != settings_menu_pos_old)
        {
            lcd_setcursor(0,settings_menu_pos_old+1);
            lcd_writedata(); lcd_send(' ');
            lcd_setcursor(0,settings_menu_pos+1);
            lcd_writedata(); lcd_send('>');
            settings_menu_pos_old = settings_menu_pos;
        }
        if ((buttons_pressed & BTN_OK))
        {
            switch(settings_menu_pos)
            {
                default:break;

                case 0:
                    calib_menu();

                case 1: // sound setting
                    stat &= stat_inside_submenu;
                    stat |= stat_inside_menu;
                    stat2 |= stat2_redraw_menu;
                    lcd_clear();
                    return;
                    break;
            }

        }
    }
}

void calib_menu()
{
/*    lcd_clear();
    uint8_t settings_menu_pos=0, settings_menu_pos_old=0;

    lcd_print_pmem_setpos(str_cal_menu_title0, 20, 0, 0);
    lcd_print_pmem_setpos(str_cal_menu_title1, 20, 0, 0);
    lcd_print_pmem_setpos(str_cal_menu_items[0], 14, 2, 1);
    lcd_print_pmem_setpos(str_cal_menu_items[1], 14, 2, 2);

    lcd_setcursor(0,1); lcd_writedata(); lcd_send('>');

    while(1)
    {
        buttons_pressed = get_buttons();
        if ((buttons_pressed & BTN_UP) && (settings_menu_pos>0)) settings_menu_pos--;
        if ((buttons_pressed & BTN_DOWN) && (settings_menu_pos<1)) settings_menu_pos++;
        if (settings_menu_pos != settings_menu_pos_old)
        {
            lcd_setcursor(0,settings_menu_pos_old+1);
            lcd_writedata(); lcd_send(' ');
            lcd_setcursor(0,settings_menu_pos+1);
            lcd_writedata(); lcd_send('>');
            settings_menu_pos_old = settings_menu_pos;
        }
        if ((buttons_pressed & BTN_OK))
        {
            switch(settings_menu_pos)
            {
                default:break;

                case 0:
                    calib_menu();

                case 1: // sound setting
                    stat &= stat_inside_submenu;
                    stat |= stat_inside_menu;
                    stat2 |= stat2_redraw_menu;
                    lcd_clear();
                    return;
                    break;
            }

        }
    }*/
}


void disp_menu()
{

    buttons_pressed = get_buttons();
    if(menu_pos != old_menu_pos)
    {
        lcd_setcursor(0,old_menu_pos+1-menu_pos2);
        lcd_writedata();
        lcd_send(' ');
        old_menu_pos = menu_pos;
    }
    lcd_setcursor(0,menu_pos+1-menu_pos2);
    lcd_writedata();
    lcd_send('>');

    if ((buttons_pressed & BTN_UP) && menu_pos > 0) // down button - decrease menu pos.
    {
        menu_pos--;
        if (menu_pos2 > menu_pos)
        {
            menu_pos2--;
            stat2 |= stat2_redraw_menu;
        }
    }

    if  ((buttons_pressed & BTN_DOWN) && menu_pos < 3) // up button - increase menu pos.
    {
        menu_pos++;
        if (menu_pos > menu_pos2+2)
        {
            menu_pos2++;
            stat2 |= stat2_redraw_menu;
        }
    }

    if (buttons_pressed & BTN_OK) // OK button - action.
    {
        switch(menu_pos){
            default:break;
            case 3:
                stat |= stat_inside_submenu;
                settings_submenu();
                break;
            case 2: // Internal resistance measurement
                stat |= stat_inside_submenu;
                last_option = 2;
                rin_submenu();
                break;
            case 1: // Constant current load
                stat |= stat_inside_submenu;
                last_option = 1;
                cc_submenu();
                break;
            case 0: // Measure capacity
                stat |= stat_inside_submenu;
                last_option = 0;
                cap_submenu();
                break;
        }
    }

}

void print_settings_submenu(uint8_t cursor_pos)
{
    lcd_clear();

    lcd_print_pmem_setpos(str_settings_items[0], 14, 2, 0);
    lcd_print_pmem_setpos(str_settings_items[1], 14, 2, 1);
    lcd_print_pmem_setpos(str_settings_items[2], 14, 2, 2);
    lcd_print_pmem_setpos(str_settings_items[3], 14, 2, 3);
    uint8_t settings_temp;
    if(settings & settings_sound_on) settings_temp=1; else settings_temp=0;
    lcd_print_pmem_setpos(str_settings_onoff[settings_temp], 3, 17, 1);

    lcd_setcursor(0, cursor_pos); lcd_writedata(); lcd_send('>');
}

void settings_submenu()
{
    lcd_clear();
    uint8_t settings_menu_pos=0, settings_menu_pos_old=0, settings_temp;

    print_settings_submenu(settings_menu_pos);

    lcd_setcursor(0,0); lcd_writedata(); lcd_send('>');

    // sound setting
    if(settings & settings_sound_on) settings_temp=1; else settings_temp=0;
    lcd_print_pmem_setpos(str_settings_onoff[settings_temp], 3, 17, 1);


    while(1)
    {
        buttons_pressed = get_buttons();
        if ((buttons_pressed & BTN_UP) && (settings_menu_pos>0)) settings_menu_pos--;
        if ((buttons_pressed & BTN_DOWN) && (settings_menu_pos<3)) settings_menu_pos++;
        if (settings_menu_pos != settings_menu_pos_old)
        {
            lcd_setcursor(0,settings_menu_pos_old);
            lcd_writedata(); lcd_send(' ');
            lcd_setcursor(0,settings_menu_pos);
            lcd_writedata(); lcd_send('>');
            settings_menu_pos_old = settings_menu_pos;
        }
        if ((buttons_pressed & BTN_OK))
        {
            switch(settings_menu_pos)
            {
                default:break;

                case 0:
                    stat &= stat_inside_submenu;
                    stat |= stat_inside_menu;
                    stat2 |= stat2_redraw_menu;
                    lcd_clear();
                    return;
                    break;

                case 1: // sound setting
                    settings ^= settings_sound_on;
                    if(settings & settings_sound_on) settings_temp=1; else settings_temp=0;
                    lcd_print_pmem_setpos(str_settings_onoff[settings_temp], 3, 17, 1);
                    break;

                case 2: // offset comp setting
                    set_offset_comp();
                    print_settings_submenu(settings_menu_pos);
                    break;

                case 3:
                    save_eeprom();
                    break;
            }

        }
    }
}


void rin_submenu()
{
    lcd_clear();
    lcd_print_pmem_setpos(str_rin_subm1, 20, 0, 0);
    lcd_print_pmem_setpos(str_rin_subm2, 14, 0, 1);
    current_string(buf, i1);
    lcd_sendstring(buf, 5);
    lcd_print_pmem_setpos(str_rin_subm3, 14, 0, 2);
    current_string(buf, i2);
    lcd_sendstring(buf, 5);
    lcd_print_pmem_setpos(str_rin_subm4, 20, 0, 3);

    while(1)
    {
        buttons_pressed = get_buttons();
        if (buttons_pressed & BTN_OK)
        {
            break;
        }
        else if (buttons_pressed & BTN_ESC)
        {
            escape_to_menu();
            return;
        }
        else if ((buttons_pressed & BTN_DOWN) || (buttons_pressed & BTN_UP))
        {
            if ((buttons_pressed & BTN_UP) && rin_current_choice < 9) rin_current_choice++;
            if ((buttons_pressed & BTN_DOWN) && rin_current_choice > 0) rin_current_choice--;
            i1 = pgm_read_byte_near(rin_current1+rin_current_choice);
            i2 = pgm_read_byte_near(rin_current2+rin_current_choice);
            lcd_setcursor(14, 1);
            current_string(buf, i1);
            lcd_sendstring(buf, 5);
            lcd_setcursor(14, 2);
            current_string(buf, i2);
            lcd_sendstring(buf, 5);
        }
    }

    stat &= stat_inside_submenu;
    stat &= stat_inside_menu;

    rin_test();
}

void cc_submenu()
{
    lcd_clear();
    lcd_print_pmem_setpos(str_ccl_subm1, 20, 0, 0);
    lcd_print_pmem_setpos(str_current, 9, 0, 1);
    current_string(buf, pwm_current);
    lcd_sendstring(buf, 5);

    submenu_stat = 1;

    uint8_t xpos;

    lcd_writecmd();
    lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY | LCD_ONOFF_BLINK);

    lcd_setcursor(12, 1);

    while(1)
    {
        buttons_pressed = get_buttons();
        if (buttons_pressed & BTN_ESC)
        {
            submenu_stat--;
            if (submenu_stat == 0)
            {
                escape_to_menu();
                break;
            }

            if (submenu_stat>0 && submenu_stat<4) // current field
            {
                xpos = 13 - submenu_stat;
                if (submenu_stat >= 3) xpos--;
                lcd_setcursor(xpos, 1);
            }

        }
        if (buttons_pressed & BTN_OK)
        {
            submenu_stat++;
            if (submenu_stat>=0 && submenu_stat<4) // current field
            {
                xpos = 13 - submenu_stat;
                if (submenu_stat >= 3) xpos--;
                lcd_setcursor(xpos, 1);
            }
            if (submenu_stat >= 4)
            {
                lcd_writecmd();
                lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY);
                lcd_clear();
                lcd_print_pmem_setpos(str_esc_exit, 20, 0, 3); // escape command
                stat &= ~(stat_inside_submenu | stat_inside_menu);
                stat |= stat_ccload_enabled;
                enable_wdt();
                set_current(pwm_current);
                enable_1s_timer();
                return;
            }
        }
        if ((buttons_pressed & BTN_UP) || (buttons_pressed & BTN_DOWN))
        {
            if (buttons_pressed & BTN_UP)
            {
                if(submenu_stat==3 && pwm_current < 156) pwm_current += 100;
                if(submenu_stat==2 && pwm_current < 246) pwm_current += 10;
                if(submenu_stat==1 && pwm_current < 255) pwm_current += 1;

            }
            if (buttons_pressed & BTN_DOWN)
            {
                if(submenu_stat==3 && pwm_current > 100) pwm_current -= 100;
                if(submenu_stat==2 && pwm_current > 10) pwm_current -= 10;
                if(submenu_stat==1 && pwm_current > 1) pwm_current -= 1;
            }
            lcd_setcursor(9,1);
            current_string(buf, pwm_current);
            lcd_sendstring(buf, 5);
            xpos = 13 - submenu_stat;
            if (submenu_stat == 3) xpos--;
            lcd_setcursor(xpos, 1);
        }

    }
}

void cap_submenu()
{
    lcd_clear();
    lcd_print_pmem_setpos(str_submenu_disch, 20, 0, 0);
    lcd_print_pmem_setpos(str_current, 9, 0, 1);
    current_string(buf, pwm_current);
    lcd_sendstring(buf, 5);
    lcd_print_pmem_setpos(str_term_v, 9, 0, 2);
    voltage_string(buf, term_voltage);
    lcd_sendstring(buf, 6);
    lcd_print_pmem_setpos(str_back_next, 20, 0, 3);

    lcd_writecmd();
    lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY | LCD_ONOFF_BLINK);

    lcd_setcursor(12,1);

    field = 0;

    submenu_stat = 1;

    uint8_t xpos=0;

    while(1)
    {
        buttons_pressed = get_buttons();
        if (buttons_pressed & BTN_ESC)
        {
            submenu_stat--;
            if (submenu_stat == 0)
            {
                escape_to_menu();
                break;
            }

            if (submenu_stat>0 && submenu_stat<4) // current field
            {
                field = 0;
                xpos = 13 - submenu_stat;
                if (submenu_stat >= 3) xpos--;
                lcd_setcursor(xpos, field+1);
            }

            if (submenu_stat>3 && submenu_stat<8) // voltage field
            {
                field = 1;
                xpos = 17 - submenu_stat;
                if (submenu_stat >= 6) xpos--;
                lcd_setcursor(xpos, field+1);
            }

        }
        if (buttons_pressed & BTN_OK)
        {
            submenu_stat++;
            if (submenu_stat>=0 && submenu_stat<4) // current field
            {
                field = 0;
                xpos = 13 - submenu_stat;
                if (submenu_stat >= 3) xpos--;
                lcd_setcursor(xpos, field+1);
            }

            if (submenu_stat>3 && submenu_stat<8) // voltage field
            {
                field = 1;
                xpos = 17 - submenu_stat;
                if (submenu_stat >= 6) xpos--;
                lcd_setcursor(xpos, field+1);
            }
            if (submenu_stat >= 8) // start discharge
            {
                mWsx10 = 0;
                clear_time();
                _delay_ms(8);
                lcd_writecmd();
                lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY);
                lcd_clear();
                stat &= ~(stat_inside_submenu | stat_inside_menu);
                stat |= stat_disch_enabled;
                _delay_ms(8);
                set_current(pwm_current);
                enable_1s_timer();
                enable_wdt();
                return;
            }
        }
        if ((buttons_pressed & BTN_UP) || (buttons_pressed & BTN_DOWN))
        {
            if (buttons_pressed & BTN_UP)
            {
                if(submenu_stat==3 && pwm_current < 156) pwm_current += 100;
                if(submenu_stat==2 && pwm_current < 246) pwm_current += 10;
                if(submenu_stat==1 && pwm_current < 255) pwm_current += 1;
                if(submenu_stat==7 && term_voltage < 1047) term_voltage += 1000;
                if(submenu_stat==6 && term_voltage < 1947) term_voltage += 100;
                if(submenu_stat==5 && term_voltage < 2037) term_voltage += 10;
                if(submenu_stat==4 && term_voltage < 2046) term_voltage += 1;
            }
            if (buttons_pressed & BTN_DOWN)
            {
                if(submenu_stat==3 && pwm_current > 100) pwm_current -= 100;
                if(submenu_stat==2 && pwm_current > 10) pwm_current -= 10;
                if(submenu_stat==1 && pwm_current > 1) pwm_current -= 1;
                if(submenu_stat==7 && term_voltage > 1079) term_voltage -= 1000;
                if(submenu_stat==6 && term_voltage > 179) term_voltage -= 100;
                if(submenu_stat==5 && term_voltage > 89) term_voltage -= 10;
                if(submenu_stat==4 && term_voltage > 80) term_voltage -= 1;
            }
            if (submenu_stat < 4)
            {
                lcd_setcursor(9,1);
                current_string(buf, pwm_current);
                lcd_sendstring(buf, 5);
                xpos = 13 - submenu_stat;
                if (submenu_stat == 3) xpos--;
                lcd_setcursor(xpos, field+1);
            }
            else
            {
                lcd_setcursor(9,2);
                voltage_string(buf, term_voltage);
                lcd_sendstring(buf, 6);
                xpos = 17 - submenu_stat;
                if (submenu_stat >= 6) xpos--;
                lcd_setcursor(xpos, field+1);
            }
        }
    }
}

void escape_to_menu()
{
    disable_wdt();
    asm("WDR");
    cli();
    set_current(0);
    disable_1s_timer();
    sei();
    lcd_writecmd();
    lcd_send(LCD_ONOFF | LCD_ONOFF_DISPLAY);

    stat = 0;
    menu_pos = 1;
    menu_pos2 = 0;
    lcd_clear();
    print_menu(str_menu_items, menu_pos2);
    stat |= stat_inside_menu;
}

void print_menu(const char **items, uint8_t position)
{

    lcd_print_pmem_setpos(str_select0, 10, 5, 0);
    lcd_print_pmem_setpos(items[0+position], 18, 2, 1);
    lcd_print_pmem_setpos(items[1+position], 18, 2, 2);
    lcd_print_pmem_setpos(items[2+position], 18, 2, 3);
}

void set_offset_comp() /// offset compensation setting
{
    lcd_clear();
    lcd_print_pmem(str_offset_comp1, 20);
    lcd_print_pmem_setpos(str_offset_comp2, 20, 0, 1);

    int8_t old_off = offset_comp;
    char off_string[4] ; // display string

    while(1)
    {

        uint8_t abs_val = (offset_comp<0) ? (uint8_t)(-1*offset_comp) : (uint8_t)(offset_comp);

        if (offset_comp<0)
        {
            off_string[0]='-';
        }
        else if (offset_comp>0) off_string[0]='+';
        else off_string[0] = ' ';

        off_string[1]=dig8_2(abs_val);
        off_string[2]=dig8_1(abs_val);
        off_string[3]=dig8_0(abs_val);

        lcd_setcursor(3,2);
        lcd_writedata();
        lcd_sendstring(off_string, 4);

        uint8_t buttons_pressed = get_buttons();

        if (((buttons_pressed & BTN_DOWN) != 0) && offset_comp > -128) offset_comp--;
        else if (((buttons_pressed & BTN_UP) != 0) && offset_comp < 127) offset_comp++;
        else if ((buttons_pressed & BTN_OK) != 0) break;
        else if ((buttons_pressed & BTN_ESC) != 0)
        {
            offset_comp = old_off;
            break;
        }
    }
}

void load_eeprom()
{
    uint8_t ee_block[2];
    eeprom_read_block(ee_block, 0, 2);
    if (ee_block[0] != 255) // only if the memory contains smth
    {
        settings = ee_block[0];
        _delay_ms(3);
        offset_comp = 255 & ee_block[1];
        _delay_ms(3);
    }
    else
    {
        offset_comp = DEFAULT_OFFSET_COMP;
        settings = DEFAULT_SETTINGS;
        save_eeprom();
    }
}

void save_eeprom()
{
    uint8_t old_ee_block[2];
    uint8_t new_ee_block[2];
    new_ee_block[0] = settings;
    new_ee_block[1] = 255 & offset_comp;

    eeprom_read_block(old_ee_block, 0, 2);

    if (new_ee_block[0] != old_ee_block[0] || new_ee_block[1] != old_ee_block[1])
    {
        eeprom_write_block(new_ee_block, 0, 2);
        beep_counter = 100;
    }

    else
    {
        beep_counter=2;
    }

    _delay_ms(10);
}

void print_temp_error()
{
    lcd_clear();
    lcd_print_pmem_setpos(str_thermal_error_1, 20, 0, 0);
    lcd_print_pmem_setpos(str_thermal_error_2, 20, 0, 1);
}

void deconf_regs()
{
    TCCR0B = 0;
    TCCR0A = 0;
    OCR0A = 0;
    TCCR1B = 0;
    OCR1A = 0;
    TIMSK0 = 0;
    TIMSK1 = 0;
    OCR2A = 0;
    TCCR2A = 0;
    TCCR2B = 0;
    ADMUX = 0;
    ADCSRA = 0;
    PORTB = 0;
    DDRB = 0;
    INHIBIT_DDR |= (1<<INHIBIT_PIN);
    INHIBIT_PORT |= (1<<INHIBIT_PIN);
    FAN_DDR |= (1<<FAN_PIN);
    FAN_PORT |= (1<<FAN_PIN);
}
